#include <linux/errno.h>
#include <linux/kernel.h>
#include <linux/module.h>
#include <linux/version.h>
#include <linux/if.h>
#include <linux/in.h>
#include <linux/net.h>
#include <net/sock.h>
#include <linux/file.h>
#include <linux/input.h>
#include <linux/poll.h>
#include <linux/signal.h>

#ifdef CONFIG_LOCK_KERNEL
#include <linux/smp_lock.h>
#endif

#if LINUX_VERSION_CODE >= KERNEL_VERSION(3, 7, 0)
#include <linux/kthread.h>
#endif

#include <asm/termios.h>
#include <asm/termbits.h>
#include <asm/i387.h>

#define CURRENT_SECONDS get_seconds()
#define RECALC_SIGPENDING	recalc_sigpending()
#define SIGNAL_MASK_LOCK(task, flags)                                  \
  spin_lock_irqsave(&task->sighand->siglock, flags)
#define SIGNAL_MASK_UNLOCK(task, flags)                                \
  spin_unlock_irqrestore(&task->sighand->siglock, flags)

#define LONG(x) ((x)/__BITS_PER_LONG)
#define __BIT(x) (1 << (x%__BITS_PER_LONG))

#define READ_INTERVAL	7

struct tb {
	struct file *dev;
	struct input_dev *input;
	struct socket* socket_comm;
	struct timer_list device_timer;
	wait_queue_head_t wq;
	int shutdown;
	int abort;
};

static struct tb tb = { 0 };

extern int ReadInput(void);
extern int InitComm(void);
extern void SetPosition(unsigned char *data, int count);
extern void notify(unsigned char type, unsigned char value);

#define DEBUG

void debug(const char *fmt, ...)
{
#ifdef DEBUG
	va_list args;

	va_start(args, fmt);
	vprintk(fmt, args);
	va_end(args);
#endif
	return;
}

void sleep(int msec)
{
	DEFINE_WAIT(wait);

	prepare_to_wait(&tb.wq, &wait, TASK_INTERRUPTIBLE);
	schedule_timeout(msec * HZ / 1000);
	finish_wait(&tb.wq, &wait);
}

void local_kernel_fpu_begin(void)
{
	kernel_fpu_begin();
}
void local_kernel_fpu_end(void)
{
	kernel_fpu_end();
}

int load_position(void)
{
	struct file *file;
	int rc = 0;
	loff_t size, pos;
	unsigned char *data;
        mm_segment_t   oldmm = get_fs();
	
	file = filp_open("/var/.pointposition", O_RDWR, 0644);
	if (IS_ERR(file)) {
		rc = PTR_ERR(file);
		debug(KERN_ERR"cannot open position file: rc = %d\n", rc);
		return rc;
	}

	pos = 0;
	size = i_size_read(file->f_dentry->d_inode);
	if (size > 128)
		size = 128;

	data = kmalloc(size, GFP_KERNEL);
	if (data == NULL) {
		filp_close(file, 0);
		return -ENOMEM;
	}

	set_fs(KERNEL_DS);
	vfs_read(file, data, size, &pos);
	set_fs(oldmm);

	SetPosition(data, size);

	kfree(data);
	filp_close(file, 0);
	return rc;
}

int save_position(unsigned char *data, int count)
{
	struct file *file;
	int rc = 0;
	loff_t pos;
        mm_segment_t   oldmm = get_fs();
	
	file = filp_open("/var/.pointposition", O_RDWR | O_CREAT, 0644);
	if (IS_ERR(file)) {
		rc = PTR_ERR(file);
		debug(KERN_ERR"cannot open position file: rc = %d\n", rc);
		return rc;
	}

	pos = 0;
	set_fs(KERNEL_DS);
	rc = vfs_write(file, data, count, &pos);
	set_fs(oldmm);

	filp_close(file, 0);

	return rc;
}

static int sync(unsigned char value)
{
	loff_t pos = 0;
	int rc = 0;
	unsigned char sync = value;
        mm_segment_t   oldmm = get_fs();
	
	if (tb.dev == NULL)
		return -ENODEV;

	set_fs(KERNEL_DS);
	rc = vfs_write(tb.dev, &value, 1, &pos);
	set_fs(oldmm);
}

static int poll_one(struct file *file, int timeout)
{
	DEFINE_WAIT(wait);
	int retval = 1;
	poll_table *table;
	struct poll_wqueues pwq;
	long to = jiffies + timeout*HZ;
	
	poll_initwait(&pwq);
	table = &pwq.pt;
	for (;;) {
		int mask;

		mask = file->f_op->poll(file, table);
		if (mask & (POLLERR | POLLHUP)) {
			retval = -1;
			break;
		}

		if (mask & POLLIN)
			break;

		if (tb.shutdown)
			break;

		if (signal_pending(current)) {
			retval = -ERESTARTSYS;
			break;
		}

		if (tb.shutdown) {
			retval = -ESHUTDOWN;
			break;
		}

		prepare_to_wait(&tb.wq, &wait, TASK_INTERRUPTIBLE);
		schedule_timeout(HZ/8);
		finish_wait(&tb.wq, &wait);

		if (timeout > 0 && jiffies > to) {
			retval = -ETIMEDOUT;
			break;
		}
	}

	set_current_state(TASK_RUNNING);
	poll_freewait(&pwq);

	return retval;
}

int sendto_device(unsigned char *data, int count)
{
	int rc;
	loff_t pos = 0;
        mm_segment_t   oldmm = get_fs();

	if (tb.dev == NULL)
		return -ENODEV;

	if (count <= 0 || count > 32)
		return -EINVAL;

	set_fs(KERNEL_DS);
	rc = vfs_write(tb.dev, data, count, &pos);
	set_fs(oldmm);

	return rc;
}

int readfrom_device(unsigned char *data, int count, int timeout)
{
	int rc = 0;
	loff_t pos;
	int size = 0;
        mm_segment_t   oldmm = get_fs();

	if (tb.dev == NULL)
		return -ENODEV;

	if (count <= 0 || count > 32)
		return -EINVAL;

	while (count > 0) {
		rc = poll_one(tb.dev, READ_INTERVAL);
		if (rc < 0) {
			if (rc != -ETIMEDOUT)
				debug(KERN_ERR"poll_one failed %d\n", rc);

			return rc;
		}

		if (tb.shutdown)
			return size;

		pos = tb.dev->f_pos;
		set_fs(KERNEL_DS);
		rc = vfs_read(tb.dev, data, count, &pos);
		set_fs(oldmm);
		tb.dev->f_pos = pos;

		if (rc < 0) {
			debug(KERN_ERR"vfs_read failed %d\n", rc);
			return rc;
		}

		data += rc;
		size += rc;
		count -= rc;
	}

#if 0
	sprintf(szBuf, "%s(%d)", KERN_ERR, size);
	for (i=0; i<size; i++) {
		int value;
		char str[8];
		value = orig_buf[i];
		sprintf(str, " %x", value);
		strcat(szBuf, str);
	}
	debug(szBuf);
#endif
	return size;
}

static int sock_write(struct socket *sock, int destport, void *buffer, 
                      int nob, int timeout)
{
        int            rc;
        struct sockaddr_in  locaddr;
        mm_segment_t   oldmm = get_fs();
        long           ticks = timeout * HZ;
        unsigned long  then;
        struct timeval tv;
	
        memset(&locaddr, 0, sizeof(locaddr));
        locaddr.sin_family = AF_INET;
        locaddr.sin_port = htons(destport);
        locaddr.sin_addr.s_addr = INADDR_ANY;

        /* Caller may pass a zero timeout if she thinks the socket buffer is
         * empty enough to take the whole message immediately */

        for (;;) {
                struct iovec  iov = {
                        .iov_base = buffer,
                        .iov_len  = nob
                };
                struct msghdr msg = {
                        .msg_name       = &locaddr,
                        .msg_namelen    = sizeof(locaddr),
                        .msg_iov        = &iov,
                        .msg_iovlen     = 1,
                        .msg_control    = NULL,
                        .msg_controllen = 0,
                        .msg_flags      = (timeout == 0) ? MSG_DONTWAIT : 0
                };

                if (timeout != 0) {
                        /* Set send timeout to remaining time */
                        tv = (struct timeval) {
                                .tv_sec = ticks / HZ,
                                .tv_usec = ((ticks % HZ) * 1000000) / HZ
                        };
                        set_fs(KERNEL_DS);
                        rc = sock_setsockopt(sock, SOL_SOCKET, SO_SNDTIMEO,
                                             (char *)&tv, sizeof(tv));
                        set_fs(oldmm);
                        if (rc != 0) {
                                debug(KERN_ERR"Can't set socket send timeout "
                                       "%ld.%06d: %d\n",
                                       (long)tv.tv_sec, (int)tv.tv_usec, rc);
                                return rc;
                        }
                }

                set_fs (KERNEL_DS);
                then = jiffies;
                rc = sock_sendmsg (sock, &msg, iov.iov_len);
                ticks -= jiffies - then;
                set_fs (oldmm);

                if (rc == nob)
                        return 0;

                if (rc < 0)
                        return rc;

                if (rc == 0) {
                        debug(KERN_ERR"Unexpected zero rc\n");
                        return (-ECONNABORTED);
                }

                if (ticks <= 0)
                        return -EAGAIN;

                buffer = ((char *)buffer) + rc;
                nob -= rc;
        }

        return (0);
}

static int sock_read(struct socket *sock, void *buffer, int nob, int timeout)
{
        int            rc;
        mm_segment_t   oldmm = get_fs();
        long           ticks = timeout * HZ;
        unsigned long  then;
        struct timeval tv;

        for (;;) {
                struct iovec  iov = {
                        .iov_base = buffer,
                        .iov_len  = nob
                };
                struct msghdr msg = {
                        .msg_name       = NULL,
                        .msg_namelen    = 0,
                        .msg_iov        = &iov,
                        .msg_iovlen     = 1,
                        .msg_control    = NULL,
                        .msg_controllen = 0,
                        .msg_flags      = 0
                };

                /* Set receive timeout to remaining time */
                tv = (struct timeval) {
                        .tv_sec = ticks / HZ,
                        .tv_usec = ((ticks % HZ) * 1000000) / HZ
                };
                set_fs(KERNEL_DS);
                rc = sock_setsockopt(sock, SOL_SOCKET, SO_RCVTIMEO,
                                     (char *)&tv, sizeof(tv));
                set_fs(oldmm);
                if (rc != 0) {
                        debug(KERN_ERR"Can't set socket recv timeout "
                               "%ld.%06d: %d\n",
                               (long)tv.tv_sec, (int)tv.tv_usec, rc);
                        return rc;
                }

                set_fs(KERNEL_DS);
                then = jiffies;
                rc = sock_recvmsg(sock, &msg, iov.iov_len, 0);
                ticks -= jiffies - then;
                set_fs(oldmm);

                if (rc < 0)
                        return rc;

                if (rc == 0)
                        return -ECONNRESET;

                buffer = ((char *)buffer) + rc;
                nob -= rc;

                if (nob == 0)
                        return 0;

                if (ticks <= 0)
                        return -ETIMEDOUT;
        }
}

int sendmsg(int port, unsigned char *data, int count)
{
	return sock_write(tb.socket_comm, port, data, count, 1);
}

int readmsg(unsigned char *data, int count)
{
	return sock_read(tb.socket_comm, data, count, 1);
}

static int local_sock_create(struct socket **sockp, int *fatal, __u32 local_ip,
                       int local_port)
{
        struct sockaddr_in  locaddr;
        struct socket      *sock;
        int                 rc;
        int                 option;
        mm_segment_t        oldmm = get_fs();

        /* All errors are fatal except bind failure if the port is in use */
        *fatal = 1;

        rc = sock_create(PF_INET, SOCK_DGRAM, 0, &sock);
        *sockp = sock;
        if (rc != 0) {
                debug(KERN_ERR"Can't create socket: %d\n", rc);
                return (rc);
        }

        set_fs (KERNEL_DS);
        option = 1;
        rc = sock_setsockopt(sock, SOL_SOCKET, SO_REUSEADDR,
                             (char *)&option, sizeof (option));
        set_fs (oldmm);
        if (rc != 0) {
                debug(KERN_ERR"Can't set SO_REUSEADDR for socket: %d\n", rc);
                goto failed;
        }

        if (local_ip != 0 || local_port != 0) {
                memset(&locaddr, 0, sizeof(locaddr));
                locaddr.sin_family = AF_INET;
                locaddr.sin_port = htons(local_port);
                locaddr.sin_addr.s_addr = (local_ip == 0) ?
                                          INADDR_ANY : htonl(local_ip);

                rc = sock->ops->bind(sock, (struct sockaddr *)&locaddr,
                                     sizeof(locaddr));
                if (rc == -EADDRINUSE) {
                        debug(KERN_ERR"Port %d already in use\n", local_port);
                        *fatal = 0;
                        goto failed;
                }
                if (rc != 0) {
                        debug(KERN_ERR"Error trying to bind to port %d: %d\n",
                               local_port, rc);
                        goto failed;
                }
        }

        return 0;

 failed:
        sock_release(sock);
        return rc;
}

#if 0
static int sock_accept (struct socket **newsockp, struct socket *sock)
{
        wait_queue_t   wait;
        struct socket *newsock;
        int            rc;

        init_waitqueue_entry(&wait, current);

        /* XXX this should add a ref to sock->ops->owner, if
         * TCP could be a module */
        rc = sock_create_lite(PF_PACKET, sock->type, IPPROTO_TCP, &newsock);
        if (rc) {
                debug(KERN_ERR"Can't allocate socket\n");
                return rc;
        }

        newsock->ops = sock->ops;

        set_current_state(TASK_INTERRUPTIBLE);
        add_wait_queue(sock->sk->sk_sleep, &wait);

        rc = sock->ops->accept(sock, newsock, O_NONBLOCK);
        if (rc == -EAGAIN) {
                /* Nothing ready, so wait for activity */
                schedule();
                rc = sock->ops->accept(sock, newsock, O_NONBLOCK);
        }

        remove_wait_queue(sock->sk->sk_sleep, &wait);
        set_current_state(TASK_RUNNING);

        if (rc != 0)
                goto failed;

        *newsockp = newsock;
        return 0;

 failed:
        sock_release(newsock);
        return rc;
}
#endif

#if LINUX_VERSION_CODE < KERNEL_VERSION(3, 7, 0)
static void __daemonize(char *str) {
        unsigned long flags;

#ifdef CONFIG_LOCK_KERNEL
        lock_kernel();
#endif

        daemonize(str);
        SIGNAL_MASK_LOCK(current, flags);
        sigfillset(&current->blocked);
        RECALC_SIGPENDING;
        SIGNAL_MASK_UNLOCK(current, flags);

#ifdef CONFIG_LOCK_KERNEL
        unlock_kernel();
#endif
}

static sigset_t block_allsigs(void)
{
        unsigned long          flags;
        sigset_t        old;

        SIGNAL_MASK_LOCK(current, flags);
        old = current->blocked;
        sigfillset(&current->blocked);
        RECALC_SIGPENDING;
        SIGNAL_MASK_UNLOCK(current, flags);

        return old;
}
#endif

static int module_reference = 0;
static int comm(void *arg)
{
	unsigned char data[4];
	int rc;

	module_reference++;

#if LINUX_VERSION_CODE < KERNEL_VERSION(3, 7, 0)
	__daemonize("BoardComm");
	block_allsigs();
#endif

	while (tb.shutdown == 0) {
		rc = sock_read(tb.socket_comm, &data, 4, 1);
		if (rc != 0) {
			if (rc != -EAGAIN)
				debug(KERN_ERR"read from socket %d\n",rc);

			continue;
		} else {
			if (data[0] == 0xAA && data[1] == 0x55) {
				int val = (int)data[2];
				int val2 = (int)data[3];
				notify(data[2], data[3]);
				printk(KERN_ERR"type=%d, val = %d\n", val, val2);
			}
		}
	}

	sock_release(tb.socket_comm);
	module_reference--;
	return 0;
}

static int init_comm(void)
{
	int rc, fatal;

	rc = local_sock_create(&tb.socket_comm, &fatal, 0, 5599);
	if (rc != 0) {
		debug(KERN_ERR"can't create socket: port 5599 in use\n");
		return rc;
	}

#if LINUX_VERSION_CODE < KERNEL_VERSION(3, 7, 0)
	kernel_thread(comm, &tb, CLONE_VM | CLONE_FS | CLONE_FILES);
#else
	kthread_run(comm, &tb, "BoardComm");
#endif

#if 0
	rc = sock->ops->listen(sock, 1);
	if (rc != 0) {
		debug(KERN_ERR"can't listen on port 5599\n");
		sock_release(sock);
		return rc;
	}

	while (!shutdown) {
		if (socket_comm_valid) {
			schedule_timeout(2*HZ);
			continue;
		}

		rc = sock_accept(&socket_comm, sock);
		if (rc != 0)
			continue;

		if (shutdown) {
			close(socket_comm);
			break;
		}

		socket_comm_valid = 1;
		continue;
	}

	close(sock);
#endif
	return 0;
}

static int mouse_first_left_x = 0;
static int mouse_first_left_y = 0;
static int mouse_left_repeat_count = 0;
static int mouse_left_down = 0;
static int mouse_right_button_sent = 0;
static int mouse_ignore_right_button = 0;
void send_key(int type, int down)
{
	if (type == 0) {
		if (down) {
			mouse_first_left_x = 0;
			mouse_first_left_y = 0;
			mouse_left_repeat_count = 0;
			mouse_right_button_sent = 0;

			mouse_left_down = 1;
		} else if (mouse_right_button_sent) {
			mouse_right_button_sent = 0;
			mouse_left_down = 0;
			
			type = 1;
		}

		mouse_ignore_right_button = 0;
	}

	if (type == 0) {
		if (down)
			input_report_key(tb.input, BTN_LEFT, 1);
		else
			input_report_key(tb.input, BTN_LEFT, 0);
	} else if (type == 1) {
		if (down)
			input_report_key(tb.input, BTN_RIGHT, 1);
		else
			input_report_key(tb.input, BTN_RIGHT, 0);
	}
}

void send_mouse(int x, int y)
{
	if (mouse_left_down && mouse_right_button_sent == 0 && mouse_ignore_right_button == 0) {
		int diff_x;
		int diff_y;

		if (mouse_first_left_x == 0 && mouse_first_left_y == 0) {
			mouse_first_left_x = x;
			mouse_first_left_y = y;
		}

		diff_x = mouse_first_left_x - x;
		diff_y = mouse_first_left_y - y;

		debug(KERN_ERR"diff_x %d, diff_y %d, x %d y %d, first_x %d, first_y %d\n", diff_x, diff_y, x, y, mouse_first_left_x, mouse_first_left_y);
		if (diff_x < 25 && diff_x > -25 && diff_y < 25 && diff_y > -25)
			mouse_left_repeat_count++;
		else {
			mouse_left_repeat_count = 0;
			mouse_ignore_right_button = 1;
		}

		/* send right mouse */
		if (mouse_right_button_sent == 0 && mouse_left_repeat_count > 40) {
			mouse_right_button_sent = 1;

			input_report_key(tb.input, BTN_LEFT, 0);
			input_report_abs(tb.input, ABS_X, x + 1);
			input_report_abs(tb.input, ABS_Y, y + 1);
			input_sync(tb.input);	

			schedule_timeout(HZ/2);
			input_report_key(tb.input, BTN_RIGHT, 1);
		}
	}

	input_report_abs(tb.input, ABS_X, x);
	input_report_abs(tb.input, ABS_Y, y);
	input_sync(tb.input);	
}

static int read(void *arg)
{
	int rc = 0;
#if LINUX_VERSION_CODE < KERNEL_VERSION(3, 7, 0)
	__daemonize("BoardService");
	block_allsigs();
#endif
	
	if (IS_ERR(tb.dev))
		return -ENODEV;

	module_reference++;
	while (tb.shutdown == 0) {
		if (tb.abort) {
			tb.abort = 0;
			break;
		}

		rc = poll_one(tb.dev, READ_INTERVAL);
		if (rc == -ETIMEDOUT) {
			sync(0xCA);
			continue;
		} else if (rc < 0) {
			break;
		}

		rc = ReadInput();
		if (rc == -ETIMEDOUT) {
			sync(0xCA);
			continue;
		} else if (rc < 0) {
			break;
		}
	}

	filp_close(tb.dev, 0);
	/* continue to detect */
	tb.dev = NULL;
	module_reference--;
	return 0;
}

static int do_ioctl(struct file *filp, unsigned int cmd, unsigned long arg)
{
	int error = -ENOTTY;
        mm_segment_t   oldmm = get_fs();

	if (!filp->f_op)
		goto out;

	set_fs(KERNEL_DS);
	if (filp->f_op->unlocked_ioctl) {
		error = filp->f_op->unlocked_ioctl(filp, cmd, arg);
		if (error == -ENOIOCTLCMD)
			error = -EINVAL;
#if 0
	} else if (filp->f_op->ioctl) {
#ifdef CONFIG_LOCK_KERNEL
		lock_kernel();
#endif
		error = filp->f_op->ioctl(filp->f_dentry->d_inode, 
					  filp, cmd, arg);
#ifdef CONFIG_LOCK_KERNEL
		unlock_kernel();
#endif
#endif
	}

out:
	set_fs(oldmm);
	if (error != 0)
		debug(KERN_ERR"ioctl return %d\n", error);
	return error;
}

static int up_time = 0;
static int device_detect(void *arg)
{
	DEFINE_WAIT(wait);
	struct termio termio = { 0 };
	struct tb *dev = (struct tb*)arg;
        mm_segment_t   oldmm = get_fs();
	int rc = 0;

	module_reference++;

#if LINUX_VERSION_CODE < KERNEL_VERSION(3, 7, 0)
	__daemonize("detect-device");
	block_allsigs();
#endif
	
	while (tb.shutdown == 0) {
		if (tb.dev != NULL) {
			loff_t pos = 0;
			unsigned char sync = 0xCC;
			set_fs(KERNEL_DS);
			rc = vfs_write(tb.dev, &sync, 1, &pos);
			set_fs(oldmm);

			prepare_to_wait(&tb.wq, &wait, TASK_INTERRUPTIBLE);
			schedule_timeout(2*HZ);
			finish_wait(&tb.wq, &wait);

			up_time += 2;
			if (up_time == 16)
				tb.abort = 1;
			continue;
		}

		tb.dev = filp_open("/dev/ttyUSB0", O_RDWR, 0644);
		if (IS_ERR(tb.dev) && PTR_ERR(tb.dev) == -ENOENT)
			tb.dev = filp_open("/dev/ttyUSB1", O_RDWR, 0644);
		if (IS_ERR(tb.dev) && PTR_ERR(tb.dev) == -ENOENT)
			tb.dev = filp_open("/dev/ttyUSB2", O_RDWR, 0644);

		if (IS_ERR(tb.dev)) {
			rc = PTR_ERR(tb.dev);
			tb.dev = NULL;
			if (rc != -ENOENT)
				debug(KERN_ERR"cannot open device file: "
				       "rc = %d\n", rc);

			prepare_to_wait(&tb.wq, &wait, 
					TASK_INTERRUPTIBLE);
			schedule_timeout(3*HZ);
			finish_wait(&tb.wq, &wait);
			continue;
		}

		termio.c_cflag = B9600 | CS8;
		do_ioctl(tb.dev, TCSETAW, (unsigned long)&termio);

		if (InitComm() < 0) {
			schedule_timeout(2*HZ);

			termio.c_cflag = B57600 | CS8;
			do_ioctl(tb.dev, TCSETAW, (unsigned long)&termio);
			if (InitComm() < 0) {
				filp_close(tb.dev, 0);
				tb.dev = NULL;

				schedule_timeout(3*HZ);
				continue;
			}
		}

#if LINUX_VERSION_CODE < KERNEL_VERSION(3, 7, 0)
		kernel_thread(read, dev, CLONE_VM | CLONE_FS | CLONE_FILES);
#else
		kthread_run(read, dev, "BoardService");
#endif
	}

	module_reference--;
	return 0;
}

extern void sqrt_test(void);
static int __init board_init(void)
{
	if (!(tb.input = input_allocate_device()))
		return -ENOMEM;	

	tb.input->name = "BoardDriver";
	tb.input->phys = "BoardDriver/input0";
	tb.input->id.bustype = BUS_USB;
	tb.input->id.vendor = 0;
	tb.input->id.product = 0x0001;
	tb.input->id.version = 0x0001;

	tb.input->evbit[0] = __BIT(EV_KEY) | __BIT(EV_ABS) | __BIT(EV_SYN);
	tb.input->absbit[0] = __BIT(ABS_X) | __BIT(ABS_Y);
	tb.input->keybit[LONG(BTN_LEFT)] = __BIT(BTN_LEFT) | 
						  __BIT(BTN_RIGHT);

	input_set_abs_params(tb.input, ABS_X, 0, 4095, 0, 0);
	input_set_abs_params(tb.input, ABS_Y, 0, 4095, 0, 0);

	input_register_device(tb.input);

	init_waitqueue_head(&tb.wq);

#if LINUX_VERSION_CODE < KERNEL_VERSION(3, 7, 0)
	kernel_thread(device_detect, &tb, CLONE_VM | CLONE_FS |
		      CLONE_FILES);
#else
	kthread_run(device_detect, &tb, "device-detect");
#endif

	/* startup comm thread */
	init_comm();
	return 0;
}

static void __exit board_exit(void)
{
	tb.shutdown = 1;

	wake_up_all(&tb.wq);

	while (module_reference > 0) {
		set_current_state(TASK_INTERRUPTIBLE);
		schedule_timeout(HZ/5);
	}

	input_unregister_device(tb.input);

	del_timer(&tb.device_timer);
}

MODULE_DESCRIPTION("BoardDriver");
MODULE_LICENSE("GPL");

module_init(board_init)
module_exit(board_exit)
